"""Custom Bocadillo commands.

Use Click to build custom commands. For documentation, see:
https://click.palletsprojects.com
"""
import click


@click.group()
def cli():
    pass


# Write your @cli.command() functions below.



from contextlib import contextmanager
from typing import List, Coroutine, Optional

from jinja2 import Environment, FileSystemLoader
from jinja2 import Template as _Template

Template = _Template

DEFAULT_TEMPLATES_DIR = "templates"


def get_templates_environment(template_dirs: List[str]):
    return Environment(
        loader=FileSystemLoader(template_dirs),
        autoescape=True,
        enable_async=True,
    )


class TemplatesMixin:
    """Provide templating capabilities to an application class."""

    def __init__(self, templates_dir: str = None, **kwargs):
        super().__init__(**kwargs)
        if templates_dir is None:
            dirs = []
        else:
            dirs = [templates_dir]
        self._templates = get_templates_environment(dirs)
        self._templates.globals.update(self.get_template_globals())

    def get_template_globals(self) -> dict:
        return {}

    @property
    def templates_dir(self) -> Optional[str]:
        """The path where templates are searched for, or `None` if not set.

        This is built from the `templates_dir` parameter.
        """
        loader: FileSystemLoader = self._templates.loader
        try:
            return loader.searchpath[0]
        except IndexError:
            return None

    @templates_dir.setter
    def templates_dir(self, templates_dir: str):
        loader: FileSystemLoader = self._templates.loader
        loader.searchpath = [templates_dir]

    def _get_template(self, name: str) -> Template:
        return self._templates.get_template(name)

    @contextmanager
    def _prevent_async_template_rendering(self):
        """If enabled, temporarily disable async template rendering."""
        # Hot fix for a bug with Jinja2's async environment, which always
        # renders asynchronously even under `render()`.
        # Example error:
        # `RuntimeError: There is no current event loop in thread [...]`
        initial = self._templates.is_async
        self._templates.is_async = False
        try:
            yield
        finally:
            self._templates.is_async = initial

    @staticmethod
    def _prepare_context(context: dict = None, **kwargs):
        if context is None:
            context = {}
        context.update(kwargs)
        return context

    async def template(
        self, name_: str, context: dict = None, **kwargs
    ) -> Coroutine:
        """Render a template asynchronously.

        Can only be used within `async` functions.

        # Parameters

        name (str):
            Name of the template, located inside `templates_dir`.
            The trailing underscore avoids collisions with a potential
            context variable named `name`.
        context (dict):
            Context variables to inject in the template.
        kwargs (dict):
            Context variables to inject in the template.
        """
        context = self._prepare_context(context, **kwargs)
        return await self._get_template(name_).render_async(context)

    def template_sync(self, name_: str, context: dict = None, **kwargs) -> str:
        """Render a template synchronously.

        See also: #API.template().
        """
        context = self._prepare_context(context, **kwargs)
        with self._prevent_async_template_rendering():
            return self._get_template(name_).render(context)

    def template_string(
        self, source: str, context: dict = None, **kwargs
    ) -> str:
        """Render a template from a string (synchronous).

        # Parameters
        source (str): a template given as a string.

        For other parameters, see #API.template().
        """
        context = self._prepare_context(context, **kwargs)
        with self._prevent_async_template_rendering():
            template = self._templates.from_string(source=source)
            return template.render(context)



import json
from typing import Any, Callable, Dict, Union

from .exceptions import UnsupportedMediaType

MediaHandler = Callable[[Any], str]


def handle_json(value: Union[dict, list]) -> str:
    """Serialize a value to a JSON string using `json.dumps()`.

    # Parameters
    value (list or dict): the value to serialize.

    # Returns
    json (str): the serialized value.
    """
    return json.dumps(value)


def handle_text(value: Any) -> str:
    """Serialize a value to plain text using `str()`.

    # Parameters
    value (any): the value to serialize.

    # Returns
    text (str): the serialized value.
    """
    return str(value)


def get_default_handlers() -> dict:
    """Return the built-in media handlers.

    # Returns
    handlers (dict): a mapping of media types to their media handler.
    """
    return {
        Media.JSON: handle_json,
        Media.PLAIN_TEXT: handle_text,
        Media.HTML: handle_text,
    }


class Media:
    """Registry of media handlers.

    # Parameters
    media_type (str):
        The media type that will be used when serializing values.
    handlers (dict):
        A mapping of media types to `(Any) -> str` callables.
        Defaults to built-in media handlers.

    # Attributes
    JSON (str): `application/json`
    PLAIN_TEXT (str): `text/plain`
    HTML (str): `text/html`
    """

    JSON = "application/json"
    PLAIN_TEXT = "text/plain"
    HTML = "text/html"

    def __init__(
        self, media_type: str, handlers: Dict[str, MediaHandler] = None
    ):
        if handlers is None:
            handlers = get_default_handlers()
        self.handlers = handlers
        self.type = media_type

    def serialize(self, value: Any, media_type: str):
        """Serialize a value using the given media type.

        # Parameters
        value (any): the value to be serialized.
        media_type (str):
            The media type of the given value. Determines which media handler
            is used.
        """
        handler = self.handlers[media_type]
        return handler(value)

    @property
    def type(self) -> str:
        """Get or set the configured media type.

        # Parameters
        media_type (str): an HTTP content type.

        # Raises
        UnsupportedMediaType :
            If no handler exists for the given media type.
        """
        return self._default_type

    @type.setter
    def type(self, media_type: str):
        """Set the configured media type."""
        if media_type not in self.handlers:
            raise UnsupportedMediaType(
                media_type, available=list(self.handlers)
            )
        self._default_type = media_type



import os
from functools import wraps
from typing import Any, Dict, List, Optional, Tuple, Type, Union, Callable

from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.middleware.wsgi import WSGIResponder
from starlette.testclient import TestClient
from starlette.websockets import WebSocketClose
from uvicorn.main import get_logger, run
from uvicorn.reloaders.statreload import StatReload

from .app_types import ASGIApp, ASGIAppInstance, WSGIApp, Scope
from .compat import call_async
from .constants import DEFAULT_CORS_CONFIG
from .error_handlers import ErrorHandler, error_to_text
from .events import EventsMixin
from .exceptions import HTTPError
from .media import Media
from .meta import DocsMeta
from .middleware import Dispatcher
from .recipes import RecipeBase
from .redirection import Redirection
from .request import Request
from .response import Response
from .routing import RoutingMixin, RouteMatch, WebSocketRouteMatch
from .staticfiles import static
from .templates import TemplatesMixin
from .websockets import WebSocket


class API(TemplatesMixin, RoutingMixin, EventsMixin, metaclass=DocsMeta):
    """The all-mighty API class.

    This class implements the [ASGI](https://asgi.readthedocs.io) protocol.

    # Example

    ```python
    >>> import bocadillo
    >>> api = bocadillo.API()
    ```

    # Parameters

    templates_dir (str):
        The name of the directory where templates are searched for,
        relative to the application entry point.
        Defaults to `"templates"`.
    static_dir (str):
        The name of the directory containing static files, relative to
        the application entry point. Set to `None` to not serve any static
        files.
        Defaults to `"static"`.
    static_root (str):
        The path prefix for static assets.
        Defaults to `"static"`.
    allowed_hosts (list of str, optional):
        A list of hosts which the server is allowed to run at.
        If the list contains `"*"`, any host is allowed.
        Defaults to `["*"]`.
    enable_cors (bool):
        If `True`, Cross Origin Resource Sharing will be configured according
        to `cors_config`. Defaults to `False`.
        See also [CORS](../guides/http/middleware.md#cors).
    cors_config (dict):
        A dictionary of CORS configuration parameters.
        Defaults to `dict(allow_origins=[], allow_methods=["GET"])`.
    enable_hsts (bool):
        If `True`, enable HSTS (HTTP Strict Transport Security) and automatically
        redirect HTTP traffic to HTTPS.
        Defaults to `False`.
        See also [HSTS](../guides/http/middleware.md#hsts).
    enable_gzip (bool):
        If `True`, enable GZip compression and automatically
        compress responses for clients that support it.
        Defaults to `False`.
        See also [GZip](../guides/http/middleware.md#gzip).
    gzip_min_size (int):
        If specified, compress only responses that
        have more bytes than the specified value.
        Defaults to `1024`.
    media_type (str):
        Determines how values given to `res.media` are serialized.
        Can be one of the supported media types.
        Defaults to `"application/json"`.
        See also [Media](../guides/http/media.md).
    """

    _error_handlers: List[Tuple[Type[Exception], ErrorHandler]]

    def __init__(
        self,
        templates_dir: str = "templates",
        static_dir: Optional[str] = "static",
        static_root: Optional[str] = "static",
        allowed_hosts: List[str] = None,
        enable_cors: bool = False,
        cors_config: dict = None,
        enable_hsts: bool = False,
        enable_gzip: bool = False,
        gzip_min_size: int = 1024,
        media_type: Optional[str] = Media.JSON,
    ):
        super().__init__(templates_dir=templates_dir)

        self._error_handlers = []
        self.add_error_handler(HTTPError, error_to_text)

        self._extra_apps: Dict[str, Any] = {}

        self.client = self.build_client()

        if static_dir is not None:
            if static_root is None:
                static_root = static_dir
            self.mount(static_root, static(static_dir))

        self._media = Media(media_type=media_type)

        self._middleware = []
        self._asgi_middleware = []

        if allowed_hosts is None:
            allowed_hosts = ["*"]
        self.add_asgi_middleware(
            TrustedHostMiddleware, allowed_hosts=allowed_hosts
        )

        if enable_cors:
            if cors_config is None:
                cors_config = {}
            cors_config = {**DEFAULT_CORS_CONFIG, **cors_config}
            self.add_asgi_middleware(CORSMiddleware, **cors_config)

        if enable_hsts:
            self.add_asgi_middleware(HTTPSRedirectMiddleware)

        if enable_gzip:
            self.add_asgi_middleware(GZipMiddleware, minimum_size=gzip_min_size)

    def build_client(self, **kwargs) -> TestClient:
        return TestClient(self, **kwargs)

    def get_template_globals(self):
        """Return global variables available to all templates.

        # Returns
        variables (dict): a mapping of variable names to their values.
        """
        return {"url_for": self.url_for}

    def mount(self, prefix: str, app: Union[ASGIApp, WSGIApp]):
        """Mount another WSGI or ASGI app at the given prefix.

        # Parameters
        prefix (str): A path prefix where the app should be mounted, e.g. `"/myapp"`.
        app: An object implementing [WSGI](https://wsgi.readthedocs.io) or [ASGI](https://asgi.readthedocs.io) protocol.
        """
        if not prefix.startswith("/"):
            prefix = "/" + prefix
        self._extra_apps[prefix] = app

    def recipe(self, recipe: RecipeBase):
        recipe.apply(self)

    @property
    def media_type(self) -> str:
        """The currently configured media type.

        When setting it to a value outside of built-in or custom media types,
        an `UnsupportedMediaType` exception is raised.
        """
        return self._media.type

    @media_type.setter
    def media_type(self, media_type: str):
        self._media.type = media_type

    @property
    def media_handlers(self) -> dict:
        """The dictionary of supported media handlers.

        You can access, edit or replace this at will.
        """
        return self._media.handlers

    @media_handlers.setter
    def media_handlers(self, media_handlers: dict):
        self._media.handlers = media_handlers

    def add_error_handler(
        self, exception_cls: Type[Exception], handler: ErrorHandler
    ):
        """Register a new error handler.

        # Parameters
        exception_cls (Exception class):
            The type of exception that should be handled.
        handler (callable):
            The actual error handler, which is called when an instance of
            `exception_cls` is caught.
            Should accept a `req`, a `res` and an `exc`.
        """
        self._error_handlers.insert(0, (exception_cls, handler))

    def error_handler(self, exception_cls: Type[Exception]):
        """Register a new error handler (decorator syntax).

        # Example
        ```python
        >>> import bocadillo
        >>> api = bocadillo.API()
        >>> @api.error_handler(KeyError)
        ... def on_key_error(req, res, exc):
        ...     pass  # perhaps set res.content and res.status_code
        ```
        """

        def wrapper(handler):
            self.add_error_handler(exception_cls, handler)
            return handler

        return wrapper

    def _find_handler(
        self, exception_cls: Type[Exception]
    ) -> Optional[ErrorHandler]:
        for cls, handler in self._error_handlers:
            if issubclass(exception_cls, cls):
                return handler
        return None

    async def _handle_exception(
        self, req: Request, res: Response, exception: Exception
    ) -> None:
        # Handle an exception raised during dispatch.
        # If no handler was registered for the exception, it is raised.
        handler = self._find_handler(exception.__class__)
        if handler is None:
            return await self._handle_exception(req, res, HTTPError(500))

        await call_async(handler, req, res, exception)
        if res.status_code is None:
            res.status_code = 500

    def redirect(
        self,
        *,
        name: str = None,
        url: str = None,
        permanent: bool = False,
        **kwargs,
    ):
        """Redirect to another route.

        # Parameters
        name (str): name of the route to redirect to.
        url (str): URL of the route to redirect to, required if `name` is omitted.
        permanent (bool):
            If `False` (the default), returns a temporary redirection (302).
            If `True`, returns a permanent redirection (301).
        kwargs (dict):
            Route parameters.

        # Raises
        Redirection: an exception that will be caught by #API.dispatch().

        # See Also
        - [Redirecting](../guides/http/redirecting.md)
        """
        if name is not None:
            url = self.url_for(name=name, **kwargs)
        else:
            assert url is not None, "url is expected if no route name is given"
        raise Redirection(url=url, permanent=permanent)

    def add_middleware(self, middleware_cls, **kwargs):
        """Register a middleware class.

        # Parameters

        middleware_cls (Middleware class):
            A subclass of `bocadillo.Middleware`.

        # See Also
        - [Middleware](../guides/http/middleware.md)
        """
        self._middleware.insert(0, (middleware_cls, kwargs))

    def add_asgi_middleware(self, middleware_cls, *args, **kwargs):
        """Register an ASGI middleware class.

        # Parameters
        middleware_cls (Middleware class):
            A class that complies with the ASGI specification.

        # See Also
        - [Middleware](../guides/http/middleware.md)
        - [ASGI](https://asgi.readthedocs.io)
        """
        self._asgi_middleware.insert(0, (middleware_cls, args, kwargs))

    def apply_asgi_middleware(self, app: ASGIApp) -> ASGIApp:
        """Wrap the registered ASGI middleware around an ASGI app.

        # Parameters
        app (ASGIApp): a callable complying with the ASGI specification.

        # Returns
        app_with_asgi_middleware (ASGIApp):
            The result `app = asgi(app, *args, **kwargs)` for
            each ASGI middleware class.

        # See Also
        - [add_asgi_middleware](#add-asgi-middleware)
        """
        for middleware_cls, args, kwargs in self._asgi_middleware:
            app: ASGIApp = middleware_cls(app, *args, **kwargs)
        return app

    async def dispatch(self, req: Request) -> Response:
        """Dispatch a request and return a response.

        # Parameters
        req (Request): an inbound HTTP request.

        # Returns
        response (Response): an HTTP response.

        # See Also
        - [How are requests processed?](../guides/http/routing.md#how-are-requests-processed) for the dispatch algorithm.
        """
        res = Response(req, media=self._media)

        try:
            match: RouteMatch = self._router.match(req.url.path)
            if match is None:
                raise HTTPError(status=404)

            try:
                await match.route(req, res, **match.params)
            except Redirection as redirection:
                res = redirection.response

        except Exception as e:
            await self._handle_exception(req, res, e)

        return res

    def _convert_exception_to_response(
        self, dispatch: Dispatcher
    ) -> Dispatcher:
        # Wrap call to `dispatch()` to always return an HTTP response.

        @wraps(dispatch)
        async def inner(req: Request) -> Response:
            try:
                res = await dispatch(req)
            except Exception as exc:
                res = Response(req, media=self._media)
                await self._handle_exception(req, res, exc)
            return res

        return inner

    async def get_response(self, req: Request) -> Response:
        """Return a response for an incoming request.

        # Parameters
        req (Request): a Request object.

        # Returns
        res (Response):
            a Response object, obtained by going down the middleware chain,
            calling `dispatch()` and going up the middleware chain.

        # See Also
        - [dispatch](#dispatch)
        - [Middleware](../guides/http/middleware.md)
        """
        convert = self._convert_exception_to_response
        dispatch = convert(self.dispatch)
        for cls, kwargs in self._middleware:
            middleware = cls(dispatch, **kwargs)
            dispatch = convert(middleware)
        return await dispatch(req)

    async def dispatch_websocket(self, scope: Scope, receive, send):
        match: WebSocketRouteMatch = self._router.match(
            scope["path"], websocket=True
        )
        if match is None:
            # Close with a 403 error code, as specified in the ASGI spec:
            # https://asgi.readthedocs.io/en/latest/specs/www.html#close
            await WebSocketClose(code=403)(receive, send)
        else:
            ws = WebSocket(scope, receive, send, **match.route.ws_kwargs)
            try:
                await match.route(ws, **match.params)
            except BaseException:
                await ws.ensure_closed(1011)
                raise

    def create_app(self, scope: Scope) -> ASGIAppInstance:
        """Build and return an instance of the `API`'s own ASGI application.

        # Parameters
        scope (dict): an ASGI connection scope.

        # Returns
        asgi (ASGIAppInstance):
            creates a `Request` and awaits the result of `get_response()`.
        """

        async def asgi(receive, send):
            nonlocal scope
            if scope["type"] == "websocket":
                await self.dispatch_websocket(scope, receive, send)
            else:
                assert scope["type"] == "http"
                req = Request(scope, receive)
                res = await self.get_response(req)
                await res(receive, send)

        return asgi

    def find_app(self, scope: Scope) -> ASGIAppInstance:
        """Return the ASGI application suited to the given ASGI scope.

        The application is chosen according to the following algorithm:

        - If `scope` has a `lifespan` type, the lifespan handler is returned.
        This occurs on server startup and shutdown.
        - If the scope's `path` begins with any of the prefixes of a mounted
        sub-app, said sub-app is returned (converting from WSGI to ASGI if
        necessary).
        - Otherwise, the `API`'s own ASGI application is returned.

        # Parameters
        scope (dict):
            An ASGI scope.

        # Returns
        app:
            An ASGI application instance.

        # See Also
        - [Lifespan Protocol](https://asgi.readthedocs.io/en/latest/specs/lifespan.html)
        - [ASGI connection scope](https://asgi.readthedocs.io/en/latest/specs/main.html#connection-scope)
        - [Events](../guides/agnostic/events.md)
        - [mount](#mount)
        - [create_app](#create-app)
        """
        if scope["type"] == "lifespan":
            return self.handle_lifespan(scope)

        path: str = scope["path"]

        # Return a sub-mounted extra app, if found
        for prefix, app in self._extra_apps.items():
            if not path.startswith(prefix):
                continue
            # Remove prefix from path so that the request is made according
            # to the mounted app's point of view.
            scope["path"] = path[len(prefix) :]
            try:
                return app(scope)
            except TypeError:
                return WSGIResponder(app, scope)

        return self.apply_asgi_middleware(self.create_app)(scope)

    def run(
        self,
        host: str = None,
        port: int = None,
        debug: bool = False,
        log_level: str = "info",
        _run: Callable = None,
        **kwargs,
    ):
        """Serve the application using [uvicorn](https://www.uvicorn.org).

        # Parameters

        host (str):
            The host to bind to.
            Defaults to `"127.0.0.1"` (localhost).
            If not given and `$PORT` is set, `"0.0.0.0"` will be used to
            serve to all known hosts.
        port (int):
            The port to bind to.
            Defaults to `8000` or (if set) the value of the `$PORT` environment
            variable.
        debug (bool):
            Whether to serve the application in debug mode. Defaults to `False`.
        log_level (str):
            A logging level for the debug logger. Must be a logging level
            from the `logging` module. Defaults to `"info"`.
        kwargs (dict):
            Extra keyword arguments that will be passed to the Uvicorn runner.

        # See Also
        - [Configuring host and port](../guides/api.md#configuring-host-and-port)
        - [Debug mode](../guides/api.md#debug-mode)
        - [Uvicorn settings](https://www.uvicorn.org/settings/) for all
        available keyword arguments.
        """
        if _run is None:  # pragma: no cover
            _run = run

        if "PORT" in os.environ:
            port = int(os.environ["PORT"])
            if host is None:
                host = "0.0.0.0"

        if host is None:
            host = "127.0.0.1"

        if port is None:
            port = 8000

        if debug:
            reloader = StatReload(get_logger(log_level))
            reloader.run(
                run,
                {
                    "app": self,
                    "host": host,
                    "port": port,
                    "log_level": log_level,
                    "debug": debug,
                    **kwargs,
                },
            )
        else:
            _run(self, host=host, port=port, **kwargs)

    def __call__(self, scope: Scope) -> ASGIAppInstance:
        return self.find_app(scope)



import os
from inspect import getsource

import click

from . import __version__

CUSTOM_COMMANDS_FILE_ENV_VAR = "BOCA_CUSTOM_COMMANDS_FILE"


def get_custom_commands_path() -> str:
    return os.getenv(CUSTOM_COMMANDS_FILE_ENV_VAR, "boca.py")


def get_custom_commands_script_contents() -> str:
    from .scaffold import boca

    return getsource(boca)


def create_cli() -> click.Command:
    """This is the Bocadillo CLI factory.

    ::: tip
    Use this function to obtain an instance of `boca` for programmatic use.
    :::

    # Returns
    cli (click.Command): an instance of the `boca` CLI.
    """

    version_kwargs = {
        "prog_name": "Bocadillo",
        "message": "%(prog)s v%(version)s",
    }
    version_flags = ("-v", "-V", "--version")

    @click.group(cls=FileGroup, path=get_custom_commands_path())
    @click.version_option(__version__, *version_flags, **version_kwargs)
    def cli():
        pass

    @cli.command()
    def version():
        """Show the version and exit."""
        click.echo(
            version_kwargs["message"]
            % {"prog": version_kwargs["prog_name"], "version": __version__}
        )

    @cli.command(name="init:custom")
    @click.option(
        "-d", "--directory", default="", help="Where files should be generated."
    )
    def init_custom(directory: str):
        """Generate files required to build custom commands."""
        contents = get_custom_commands_script_contents()
        path = os.path.join(directory, get_custom_commands_path())
        with open(path, "w") as f:
            f.write(contents)
        click.echo(click.style(f"Generated {path}", fg="green"))
        click.echo("Open the file and start building!")

    return cli


class FileGroup(click.Group):
    """A Click [MultiCommand] that loads a group of commands from a file.

    ::: warning
    You should not need to interact with this class directly.
    :::

    [MultiCommand]: http://click.palletsprojects.com/en/7.x/api/#click.MultiCommand

    # Parameters
    path (str):
        Path to a Python module declaring at least one Click group, from which
        extra commands will be obtained.

    """

    def __init__(self, path: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._load_group(path)

    def _load_group(self, path: str):
        ns = {}

        try:
            with open(path, "r") as f:
                code = compile(source=f.read(), filename=path, mode="exec")
        except FileNotFoundError:
            # User did not create custom commands
            return
        else:
            eval(code, ns, ns)

        groups = (val for val in ns.values() if isinstance(val, click.Group))
        group = next(groups, None)
        if group is None:
            raise ValueError(
                f"Expected at least one group in {path}, none found."
            )
        for name, cmd in group.commands.items():
            self.add_command(cmd, name)



from .api import API
from .exceptions import HTTPError, WebSocketDisconnect
from .media import Media
from .middleware import Middleware
from .recipes import Recipe
from .staticfiles import static
from .views import view
from .websockets import WebSocket

__version__ = "0.9.0"



import inspect
from functools import wraps
from typing import Callable, Dict, Union, Awaitable, Type

from .compat import call_async
from .request import Request
from .response import Response
from .routing import Route
from .views import Handler, get_handlers, View

HookFunction = Callable[[Request, Response, dict], Awaitable[None]]
HookCollection = Dict[Route, HookFunction]

BEFORE = "before"
AFTER = "after"


async def empty_hook(req: Request, res: Response, params: dict):
    pass


class Hooks:
    """Hooks manager."""

    def before(self, hook: HookFunction, *args, **kwargs):
        """Register a before hook on a handler.

        # Parameters
        hook (callable): a hook function.
        """
        return self._hook_decorator(BEFORE, hook, *args, **kwargs)

    def after(self, hook: HookFunction, *args, **kwargs):
        """Register an after hook on a handler.

        # Parameters
        hook (callable): a hook function.
        """
        return self._hook_decorator(AFTER, hook, *args, **kwargs)

    def _hook_decorator(
        self, hook_type: str, hook: HookFunction, *args, **kwargs
    ):
        # Enclose args and kwargs
        async def hook_func(req: Request, res: Response, params: dict):
            await call_async(hook, req, res, params, *args, **kwargs)

        def decorator(handler: Union[Type[View], Handler]):
            """Attach the hook to the given handler."""
            if inspect.isclass(handler):
                # Handler is a view class. Apply hook to all handlers.
                view_cls = handler
                for method, handler in get_handlers(view_cls).items():
                    setattr(view_cls, method, decorator(handler))
                return view_cls
            else:
                return _with_hook(hook_type, hook_func, handler)

        return decorator


def _with_hook(hook_type: str, func: HookFunction, handler: Handler):
    async def call_hook(args, kw):
        if len(args) == 2:
            req, res = args
        else:
            # method that has `self` as a first parameter
            req, res = args[1:3]
        assert isinstance(req, Request)
        assert isinstance(res, Response)
        await call_async(func, req, res, kw)

    if hook_type == BEFORE:

        @wraps(handler)
        async def with_hook(*args, **kwargs):
            await call_hook(args, kwargs)
            await call_async(handler, *args, **kwargs)

    else:
        assert hook_type == AFTER

        @wraps(handler)
        async def with_hook(*args, **kwargs):
            await call_async(handler, *args, **kwargs)
            await call_hook(args, kwargs)

    return with_hook


# Pre-bind to module
hooks = Hooks()
before = hooks.before
after = hooks.after



from typing import List, Callable, Awaitable, MutableMapping, Any

# ASGI
Scope = dict
Event = MutableMapping[str, Any]
Receive = Callable[[], Event]
Send = Callable[[Event], None]
ASGIAppInstance = Callable[[Receive, Send], Awaitable[None]]
ASGIApp = Callable[[Scope], ASGIAppInstance]

# WSGI
Environ = dict
StartResponse = Callable[[str, List[str]], None]
WSGIApp = Callable[[Environ, StartResponse], List[bytes]]



from http import HTTPStatus
from typing import Union, Any, List
from starlette.websockets import WebSocketDisconnect as _WebSocketDisconnect


WebSocketDisconnect = _WebSocketDisconnect


class HTTPError(Exception):
    """Raised when an HTTP error occurs.

    You can raise this within a view or an error handler to interrupt
    request processing.

    # Parameters
    status (int or HTTPStatus):
        the status code of the error.
    detail (any):
        extra detail information about the error. The exact rendering is
        determined by the configured error handler for `HTTPError`.

    # See Also
    - [HTTP response status codes (MDN web docs)](https://developer.mozilla.org/en-US/docs/Web/HTTP/Status)
    """

    def __init__(self, status: Union[int, HTTPStatus], detail: Any = ""):
        if isinstance(status, int):
            status = HTTPStatus(status)
        else:
            assert isinstance(
                status, HTTPStatus
            ), f"Expected int or HTTPStatus, got {type(status)}"
        self._status = status
        self.detail = detail

    @property
    def status_code(self) -> int:
        """Return the HTTP error's status code, e.g. `404`."""
        return self._status.value

    @property
    def status_phrase(self) -> str:
        """Return the HTTP error's status phrase, e.g. `"Not Found"`."""
        return self._status.phrase

    @property
    def title(self) -> str:
        """Return the HTTP error's title, e.g. `"404 Not Found"`."""
        return f"{self.status_code} {self.status_phrase}"

    def __str__(self):
        return self.title


class UnsupportedMediaType(Exception):
    """Raised when trying to use an unsupported media type.

    # Parameters
    media_type (str):
        the unsupported media type.
    available (list of str):
        a list of supported media types.
    """

    def __init__(self, media_type: str, available: List[str]):
        self._media_type = media_type
        self._available = available

    def __str__(self):
        return f'{self._media_type} (available: {", ".join(self._available)})'


class RouteDeclarationError(Exception):
    """Raised when a route is ill-declared."""



from json import JSONDecodeError

from starlette.requests import Request as _Request

from .exceptions import HTTPError


class Request(_Request):
    """The succulent request object."""

    async def json(self):
        """Parse the request body as JSON.

        # Returns
        json (dict): the result of `json.loads(await self.body())`.

        # Raises
        HTTPError(400): if the JSON is malformed.
        """
        try:
            return await super().json()
        except JSONDecodeError:
            raise HTTPError(400, detail="JSON is malformed.")

    async def __aiter__(self):
        async for chunk in self.stream():
            yield chunk



import inspect
from typing import AnyStr, Any, Callable, Coroutine, Optional, AsyncIterable

from starlette.background import BackgroundTask
from starlette.requests import Request
from starlette.responses import (
    Response as _Response,
    StreamingResponse as _StreamingResponse,
)

from .media import Media

BackgroundFunc = Callable[..., Coroutine]
StreamFunc = Callable[[], AsyncIterable[AnyStr]]


class Response:
    """Response builder."""

    CONTENT_ATTRS = {
        "text": Media.PLAIN_TEXT,
        "html": Media.HTML,
        "media": None,
    }

    def __init__(self, request: Request, media: Media):
        self.request = request
        self._content: AnyStr = None
        self.status_code: int = None
        self.headers = {}
        self._media = media
        self._background: BackgroundFunc = None
        self._generator: AsyncIterable[bytes] = None
        self.chunked = False

    @property
    def content(self) -> Optional[AnyStr]:
        return self._content

    @content.setter
    def content(self, content: AnyStr):
        self._content = content

    def _set_media(self, value: Any, media_type: str):
        content = self._media.serialize(value, media_type=media_type)
        self.headers["content-type"] = media_type
        self._content = content

    def __setattr__(self, key, value):
        if key in self.CONTENT_ATTRS:
            media_type = self.CONTENT_ATTRS[key] or self._media.type
            self._set_media(value, media_type=media_type)
        else:
            super().__setattr__(key, value)

    def background(self, func: BackgroundFunc, *args, **kwargs):
        """Register a coroutine function to be executed in the background."""

        async def background():
            await func(*args, **kwargs)

        self._background = background
        return func

    @property
    def background_task(self) -> Optional[BackgroundTask]:
        if self._background is not None:
            return BackgroundTask(self._background)
        return None

    def stream(self, func: StreamFunc):
        """Stream the response.

        Should be used to decorate a no-argument asynchronous generator
        function.
        """
        assert inspect.isasyncgenfunction(func)
        self._generator = func()
        return func

    async def __call__(self, receive, send):
        """Build and send the response."""
        if self.status_code is None:
            self.status_code = 200

        if self.status_code != 204:
            self.headers.setdefault("content-type", Media.PLAIN_TEXT)

        if self.chunked:
            self.headers["transfer-encoding"] = "chunked"

        if self._generator is not None:
            response_cls = _StreamingResponse
            content = self._generator
        else:
            response_cls = _Response
            content = self.content

        response = response_cls(
            content=content,
            headers=self.headers,
            status_code=self.status_code,
            background=self.background_task,
        )
        await response(receive, send)



import inspect
from functools import partial
from string import Formatter
from typing import Optional, NamedTuple, Dict, Callable, Union, Type, Any, Tuple

from parse import parse

from . import views
from .exceptions import HTTPError, RouteDeclarationError
from .meta import DocsMeta
from .request import Request
from .response import Response
from .views import View, HandlerDoesNotExist
from .websockets import WebSocket, WebSocketView


class BaseRoute:
    # Base route class.

    def __init__(self, pattern: str):
        self._pattern = pattern

    def url(self, **kwargs) -> str:
        """Return full path for the given route parameters.

        # Parameters
        kwargs (dict): route parameters.

        # Returns
        url (str):
            A full URL path obtained by formatting the route pattern with
            the provided route parameters.
        """
        return self._pattern.format(**kwargs)

    def parse(self, path: str) -> Optional[dict]:
        """Parse an URL path against the route's URL pattern.

        # Returns
        result (dict or None):
            If the URL path matches the URL pattern, this is a dictionary
            containing the route parameters, otherwise None.
        """
        result = parse(self._pattern, path)
        if result is not None:
            return result.named
        return None


class Route(BaseRoute, metaclass=DocsMeta):
    """Represents the binding of an URL pattern to a view.

    Note: as a framework user, you will not need to create routes directly.

    # Parameters
    pattern (str): an URL pattern. F-string syntax is supported for parameters.
    view (View):
        A `View` object.
    name (str):
        The route's name.
    """

    def __init__(self, pattern: str, view: View, name: str):
        super().__init__(pattern)
        self._view = view
        self._name = name

    async def __call__(self, req: Request, res: Response, **kwargs) -> None:
        try:
            await self._view(req, res, **kwargs)
        except HandlerDoesNotExist as e:
            raise HTTPError(405) from e


class WebSocketRoute(BaseRoute, metaclass=DocsMeta):
    """Represents the binding of an URL path to a WebSocket view.

    # Parameters
    pattern (str): an URL pattern.
    view (coroutine function):
        Should take as parameter a `WebSocket` object and
        any extracted route parameters.
    """

    def __init__(self, pattern: str, view: WebSocketView, **kwargs):
        super().__init__(pattern)
        self._view = view
        self.ws_kwargs = kwargs

    async def __call__(self, ws: WebSocket, **params):
        await self._view(ws, **params)


class RouteMatch(NamedTuple):
    # Represents the result of a successful route match.

    route: Route
    params: dict


class WebSocketRouteMatch(NamedTuple):
    """Represents the result of a successful websocket route match."""

    route: WebSocketRoute
    params: dict


class Router:
    # A collection of routes.

    def __init__(self):
        self._routes: Dict[str, Route] = {}
        self._websocket_routes: Dict[str, WebSocketRoute] = {}

    def add_route(
        self,
        view: Union[View, Type[Any], Callable, Any],
        pattern: str,
        *,
        name: str = None,
        namespace: str = None,
    ) -> Route:
        # Build and register a route.

        if isinstance(view, View):
            # View instance. No further conversion required.
            pass
        elif inspect.isclass(view):
            # View-like class.
            view = views.from_obj(view())
        elif callable(view):
            # Function-based view.
            # NOTE: here, we ensure backward-compatibility with the routing of
            # function-based views pre-0.9.
            view = views.from_handler(view)
        else:
            # View-like object.
            view = views.from_obj(view)

        assert isinstance(view, View)

        if name is None:
            name = view.name
        if namespace is not None:
            name = namespace + ":" + name

        check_route(pattern, view)

        route = Route(pattern=pattern, view=view, name=name)
        self._routes[name] = route

        return route

    def route(self, *args, **kwargs):
        # Register a route by decorating a view.
        return partial(self.add_route, *args, **kwargs)

    def match(self, path: str, websocket: bool = False) -> Optional[RouteMatch]:
        # Attempt to match an URL path against one of the registered routes.
        # NOTE: websocket (bool) = whether to look for an HTTP (`False`)
        # or WebSocket (`True`) route.
        if websocket:
            source = self._websocket_routes
            match_cls = WebSocketRouteMatch
        else:
            source = self._routes
            match_cls = RouteMatch

        for pattern, route in source.items():
            params = route.parse(path)
            if params is not None:
                return match_cls(route=route, params=params)

        return None

    def get_route_or_404(self, name: str) -> Route:
        # Return a route or raise a 404 error.
        try:
            return self._routes[name]
        except KeyError as e:
            raise HTTPError(404) from e

    def websocket_route(self, pattern: str, **kwargs):
        # Register a WebSocket route by decorating a view.

        def decorate(view):
            nonlocal self
            route = WebSocketRoute(pattern=pattern, view=view, **kwargs)
            self._websocket_routes[pattern] = route
            return route

        return decorate


class RoutingMixin:
    """Provide routing capabilities to a class."""

    def __init__(self):
        super().__init__()
        self._router = Router()

    def route(self, pattern: str, *, name: str = None, namespace: str = None):
        """Register a new route by decorating a view.

        # Parameters
        pattern (str): an URL pattern.
        methods (list of str):
            An optional list of HTTP methods.
            Defaults to `["get", "head"]`.
            Ignored for class-based views.
        name (str):
            An optional name for the route.
            If a route already exists for this name, it is replaced.
            Defaults to a snake-cased version of the view's name.
        namespace (str):
            An optional namespace for the route. If given, it is prefixed to
            the name and separated by a colon.

        # Raises
        RouteDeclarationError:
            If route validation has failed.

        # See Also
        - [check_route](#check-route) for the route validation algorithm.
        """
        return self._router.route(
            pattern=pattern, name=name, namespace=namespace
        )

    def websocket_route(
        self,
        pattern: str,
        *,
        value_type: Optional[str] = None,
        receive_type: Optional[str] = None,
        send_type: Optional[str] = None,
        caught_close_codes: Optional[Tuple[int]] = None,
    ):
        """Register a WebSocket route by decorating a view.

        # Parameters
        pattern (str): an URL pattern.

        # See Also
        - [WebSocket](./websockets.md#websocket) for a description of keyword
        arguments.
        """
        # NOTE: use named keyword arguments instead of `**kwargs` to improve
        # their accessibility (e.g. for IDE discovery).
        return self._router.websocket_route(
            pattern,
            value_type=value_type,
            receive_type=receive_type,
            send_type=send_type,
            caught_close_codes=caught_close_codes,
        )

    def url_for(self, name: str, **kwargs) -> str:
        """Build the URL path for a named route.

        # Parameters
        name (str): the name of the route.
        kwargs (dict): route parameters.

        # Returns
        url (str): the URL path for a route.

        # Raises
        HTTPError(404) : if no route exists for the given `name`.
        """
        route = self._router.get_route_or_404(name)
        return route.url(**kwargs)


def check_route(pattern: str, view: View) -> None:
    """Check compatibility of a route pattern and a view.

    # Parameters
    pattern (str): an URL pattern.
    view (View): a `View` object.

    # Raises
    RouteDeclarationError :
        - If `pattern` does not have a leading slash.
        - If route parameters declared in the pattern do not match those
        of any view handler, e.g. a parameter is declared in the pattern, but
        not used in the handler or vice-versa.

    # See Also
    - `ALL_HTTP_METHODS` is defined in [constants.py](./constants.md).
    """
    if not pattern.startswith("/"):
        raise RouteDeclarationError(
            f"Route pattern '{pattern}' on view '{view.name}' "
            f"must start with '/' to avoid ambiguities."
        )

    parsed_format = Formatter().parse(pattern)
    route_parameters: set = {name for _, name, _, _ in parsed_format if name}

    for method, handler in views.get_handlers(view).items():
        handler_parameters = dict(inspect.signature(handler).parameters)
        handler_parameters.pop("self", None)  # paranoia check
        handler_parameters = list(handler_parameters)

        for route_param in route_parameters:
            if route_param not in handler_parameters:
                raise RouteDeclarationError(
                    f"Route pattern '{pattern}' declares the route parameter "
                    f"'{route_param}' and should be a parameter of "
                    f"'{handler.__qualname__}'."
                )

        for handler_param in handler_parameters[2:]:
            if handler_param not in route_parameters:
                raise RouteDeclarationError(
                    f"Handler '{handler.__qualname__}' expects parameter "
                    f"'{handler_param}' but it was not declared in the route "
                    f"pattern '{pattern}'"
                )



import json
from typing import Awaitable, Callable, Optional, Any, Union, Tuple

from starlette.websockets import WebSocket as StarletteWebSocket

from .app_types import Event
from .constants import WEBSOCKET_CLOSE_CODES
from .exceptions import WebSocketDisconnect

_STARLETTE_WEBSOCKET_DOCS = (
    "[Starlette.websockets.WebSocket](https://www.starlette.io/websockets/)"
)


def _get_alias_docs(name: str) -> str:
    return f"\n\nAlias of `{name}` on {_STARLETTE_WEBSOCKET_DOCS}."


class _Delegated:
    """Descriptor to delegate a method to the underlying Starlette WebSocket.

    # See Also
    - [Descriptors](https://docs.python.org/3/reference/datamodel.html#implementing-descriptors)
    """

    def __init__(self, websocket_attr: str = "_websocket"):
        self.websocket_attr = websocket_attr
        self._docs_imported = False

    def __set_name__(self, owner, name: str):
        # Use the declared attribute's name as source attribute name.
        self.source = name

    def __get__(self, instance: Optional["WebSocket"], owner):
        if instance is None:  # pragma: no cover
            # Class attribute access.
            # NOTE: used by Pydoc-Markdown when generating docs.
            obj = getattr(StarletteWebSocket, self.source)
            if not self._docs_imported:
                obj.__doc__ = (obj.__doc__ or "") + _get_alias_docs(self.source)
                self._docs_imported = True
            return obj
        else:
            # Instance attribute access.
            return getattr(getattr(instance, self.websocket_attr), self.source)


class WebSocket:
    """Represents a WebSocket connection.

    # Parameters
    value_type (str):
        The type of messages received or sent over the WebSocket.
        If given, overrides `receive_type` and `send_type`.
        Defaults to `None`.
    receive_type (str):
        The type of messages received over the WebSocket.
        Defaults to `"text"`.
    send_type (str):
        The type of messages send over the WebSocket.
        Defaults to `"text"`.
    caught_close_codes (tuple of int):
        Close codes of `WebSocketDisconnect` exceptions that should be
        caught and silenced. Defaults to `(1000, 1001)`.
    args (any):
        Passed to the underlying Starlette `WebSocket` object. This is
        typically the ASGI `scope`, `receive` and `send` objects.
    """

    __default_receive_type__ = "text"
    __default_send_type__ = "text"

    def __init__(
        self,
        *args,
        value_type: Optional[str] = None,
        receive_type: Optional[str] = None,
        send_type: Optional[str] = None,
        caught_close_codes: Optional[Tuple[int]] = None,
    ):
        # NOTE: we use composition over inheritance here, because
        # we want to redefine `receive()` and `send()` but Starlette's
        # WebSocket class uses those in many other functions, which we
        # do not need / want to re-implement.
        # The compromise is the definition of delegated methods below.
        self._websocket = StarletteWebSocket(*args)

        if caught_close_codes is None:
            caught_close_codes = (1000, 1001)
        if caught_close_codes is all:
            caught_close_codes = list(WEBSOCKET_CLOSE_CODES)
        self.caught_close_codes = caught_close_codes

        if value_type is not None:
            receive_type = send_type = value_type
        else:
            receive_type = receive_type or self.__default_receive_type__
            send_type = send_type or self.__default_send_type__
        self.receive_type = receive_type
        self.send_type = send_type

    # Methods delegated to the underlying Starlette WebSocket object.
    # TODO: add type annotations.
    accept = _Delegated()
    close = _Delegated()
    receive_text = _Delegated()
    send_text = _Delegated()
    receive_bytes = _Delegated()
    send_bytes = _Delegated()

    async def receive_json(self) -> Union[dict, list]:
        """Return `json.loads(await self.receive_text())`.

        # Returns
        json (dict or list): a JSON message.
        """
        # Starlette decodes from bytes by default,
        # but most WebSocket clients generally send text.
        return json.loads(await self.receive_text())

    async def send_json(self, message: Union[dict, list]):
        """Execute `await self.send_text(json.dumps(message))`.

        # Parameters
        message (list or dict): a JSON message.
        """
        # Encodes as bytes by default,
        # but most WebSocket clients don't expect to receive plain bytes.
        await self.send_text(json.dumps(message))

    async def receive_event(self) -> Event:
        return await self._websocket.receive()

    receive_event.__doc__ = _get_alias_docs("receive")

    async def send_event(self, event: Event):
        return await self._websocket.send(event)

    send_event.__doc__ = _get_alias_docs("send")

    async def receive(self) -> Union[str, bytes, list, dict]:
        """Receive a message from the WebSocket.

        Shortcut for `receive_<self.receive_type>`.
        """
        receiver = getattr(self, f"receive_{self.receive_type}")
        return await receiver()

    async def send(self, message: Any):
        """Send a message over the WebSocket.

        Shortcut for `send_<self.send_type>`.
        """
        sender = getattr(self, f"send_{self.send_type}")
        return await sender(message)

    async def ensure_closed(self, code: int = 1000):
        """Close the connection if it has not been closed already.

        # Parameters
        code (int): a close code, defaults to `1000`.
        """
        try:
            await self.close(code)
        except RuntimeError:
            # Already closed.
            pass

    async def reject(self):
        """Reject the connection request.

        This is equivalent to `await close(403)`.
        Calling this before `accept()` has undefined behavior.
        """
        return await self.close(code=403)

    # Asynchronous context manager.

    async def __aenter__(self, *args, **kwargs):
        await self.accept()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type == WebSocketDisconnect:
            # Client has closed the connection.
            # Returning `True` here silences the exception. See:
            # https://docs.python.org/3/reference/datamodel.html#object.__exit__
            return exc_val.code in self.caught_close_codes
        else:
            # Close with Internal Error if an exception was raised.
            code = 1000 if exc_type is None else 1011
            # NOTE: view may have closed the WebSocket already, so we
            # must use ensure_closed().
            await self.ensure_closed(code)

    # Asynchronous iterator.

    async def __aiter__(self):
        while True:
            yield await self.receive()


WebSocketView = Callable[[WebSocket], Awaitable[None]]



ALL_HTTP_METHODS = ("GET", "HEAD", "POST", "PUT", "DELETE", "OPTIONS", "PATCH")

DEFAULT_CORS_CONFIG = {"allow_origins": [], "allow_methods": ["GET"]}

# See: https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent
WEBSOCKET_CLOSE_CODES = {
    1000: "Normal Closure",
    1001: "Going Away",
    1002: "Protocol Error",
    1003: "Unsupported Type",
    # 1004 is reserved
    1005: "No Status Code [Internal]",
    1006: "Connection Closed Abnormally [Internal]",
    1007: "Invalid Data",
    1008: "Policy Violation",
    1009: "Message Too Big",
    1010: "Extension Required",
    1011: "Internal Error",
    1015: "TLS Failure [Internal]",
}



"""Bocadillo middleware definition."""
from typing import Callable, Awaitable

from .compat import call_async
from .request import Request
from .response import Response

Dispatcher = Callable[[Request], Awaitable[Response]]


class Middleware:
    """Base class for middleware classes.

    # Parameters
    dispatch (coroutine function):
        a function whose return value can be awaited to obtain a response.
    kwargs (dict):
        Keyword arguments passed when registering the middleware.
    """

    def __init__(self, dispatch: Dispatcher, **kwargs):
        self.dispatch = dispatch
        self.kwargs = kwargs

    async def before_dispatch(self, req: Request):
        """Perform processing before a request is dispatched.

        If a `Response` object is returned, it will be used
        and no further processing will be performed.

        # Parameters
        req (Request): a Request object.
        """

    async def after_dispatch(self, req: Request, res: Response):
        """Perform processing after a request has been dispatched.

        If a `Response` object is returned, it is used instead of the response
        obtained by awaiting `dispatch()`.

        # Parameters
        req (Request): a Request object.
        res (Response): a Response object.
        """

    async def process(self, req: Request) -> Response:
        """Process an incoming request.

        Roughly equivalent to:

        ```python
        res = await self.before_dispatch(req) or None
        res = res or await self.dispatch(req)
        res = await self.after_dispatch(req, res) or res
        return res
        ```

        Aliased by `__call__()`.

        # Parameters
        req (Request): a request object.

        # Returns
        res (Response): a Response object.
        """
        res = await call_async(self.before_dispatch, req) or None
        res = res or await self.dispatch(req)
        res = await call_async(self.after_dispatch, req, res) or res
        return res

    async def __call__(self, req: Request) -> Response:
        return await self.process(req)


# TODO: remove in v0.8
RoutingMiddleware = Middleware



"""Various compatibility utilities."""
import asyncio
import re
from typing import Callable, Coroutine

from starlette.concurrency import run_in_threadpool

from .app_types import WSGIApp

_camel_regex = re.compile(r"(.)([A-Z][a-z]+)")
_snake_regex = re.compile(r"([a-z0-9])([A-Z])")


async def call_async(func: Callable, *args, sync=None, **kwargs) -> Coroutine:
    """Call a function in an async manner.

    # Parameters
    func (Callable):
        a callable that is either awaited (if a coroutine function)
        or run in the thread pool (if a regular function).
    sync (bool):
        A hint as to whether `func` is synchronous. If not given, it is
        inferred as `asyncio.iscoroutinefunction(func)`.

    # See Also
    - [Executing code in thread or process pools](https://docs.python.org/3/library/asyncio-eventloop.html#executing-code-in-thread-or-process-pools)
    """
    if sync or (sync is None and not asyncio.iscoroutinefunction(func)):
        return await run_in_threadpool(func, *args, **kwargs)
    return await func(*args, **kwargs)


def camel_to_snake(name: str) -> str:
    """Convert a `CamelCase` name to its `snake_case` version."""
    s1 = _camel_regex.sub(r"\1_\2", name)
    return _snake_regex.sub(r"\1_\2", s1).lower()


def empty_wsgi_app() -> WSGIApp:
    """Return a WSGI app that always returns 404 Not Found."""

    def wsgi(environ, start_response):
        status = "404 Not Found"
        body = b"Not Found"
        headers = [("Content-Type", "text/plain")]
        start_response(status, headers)
        return [body]

    return wsgi



from typing import List, Sequence, Tuple

from .meta import DocsMeta
from .templates import TemplatesMixin
from .websockets import WebSocketView


class RecipeRoute:
    """A specific kind of route for recipes.

    The route will be registered an actual API object during `recipe.apply()`.

    Mostly an implementation detail.
    """

    def __init__(self, pattern, view, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._pattern = pattern
        self._view = view

    def register(self, api, prefix: str) -> None:
        """Register the route on the API object at the given prefix.

        # Parameters

        api (API):
            A Bocadillo application.
        prefix (str):
            A path prefix.
        """
        api.route(prefix + self._pattern, *self._args, **self._kwargs)(
            self._view
        )


class RecipeBase:
    """Definition of the recipe interface."""

    def apply(self, api, root: str = ""):
        """Apply the recipe to an API object.

        Should be implemented by subclasses.

        # Parameters
        api (API): an API object.
        root (str): a root URL path itself prefixed to the recipe's `prefix`.
        """
        raise NotImplementedError


class Recipe(TemplatesMixin, RecipeBase, metaclass=DocsMeta):
    """A grouping of capabilities that can be merged back into an API.

    # Parameters

    name (str):
        A name for the recipe.
    prefix (str):
        The path prefix where the recipe will be mounted.
        Defaults to `"/" + name`.
    templates_dir (str):
        See #API.
    """

    def __init__(self, name: str, prefix: str = None, **kwargs):
        super().__init__(**kwargs)
        if prefix is None:
            prefix = f"/{name}"
        assert prefix.startswith("/"), "recipe prefix must start with '/'"
        self._name = name
        self._prefix = prefix
        self._routes: List[RecipeRoute] = []
        self._ws_routes: List[Tuple[str, dict, WebSocketView]] = []

    def route(self, *args, **kwargs):
        """Register a route on the recipe.

        Accepts the same arguments as `API.route()`, except `namespace` which
        is given the value of the recipe's `name`.

        # See Also
        - [API.route()](./api.md#route)
        """

        def register(view):
            route = RecipeRoute(
                *args, view=view, namespace=self._name, **kwargs
            )
            self._routes.append(route)
            return route

        return register

    def websocket_route(self, pattern: str, **kwargs):
        """Register a WebSocket route on the recipe.

        Accepts the same arguments as `API.websocket_route()`.

        # See Also
        - [API.websocket_route()](./api.md#websocket-route)
        """

        def register(view):
            self._ws_routes.append((pattern, kwargs, view))
            return view

        return register

    def apply(self, api, root: str = ""):
        """Apply the recipe to an API object.

        This will:

        - Mount registered routes onto the `api`.
        - Update the templates directory to that of `api`.

        # See Also
        - [RecipeBase.apply()](#apply)
        """
        # Apply routes on the API
        for route in self._routes:
            route.register(api, root + self._prefix)

        for pattern, kwargs, view in self._ws_routes:
            api.websocket_route(root + self._prefix + pattern, **kwargs)(view)

        # Look for templates where the API does, if not specified
        if self.templates_dir is None:
            self.templates_dir = api.templates_dir

    @classmethod
    def book(cls, *recipes: "Recipe", prefix: str) -> "RecipeBook":
        """Build a book of recipes.

        Shortcut for `RecipeBook(recipes, prefix)`.
        """
        return RecipeBook(recipes, prefix)


class RecipeBook(RecipeBase):
    """A composition of multiple recipes.

    # Parameters
    recipes (list): a list of `Recipe` objects.
    prefix (str):
        A prefix that will be prepended to all of the recipes' own prefixes.
    """

    def __init__(self, recipes: Sequence[Recipe], prefix: str):
        self.recipes = recipes
        self.prefix = prefix

    def apply(self, api, root: str = ""):
        """Apply the recipe book to an API object.

        This is equivalent to calling `recipe.apply(api, root=root + self.prefix)`
        for each of the book's recipes.

        # See Also
        - [RecipeBase.apply()](#apply)
        """
        for recipe in self.recipes:
            recipe.apply(api, root=root + self.prefix)



import inspect
from functools import partial, wraps
from typing import List, Union, Any, Awaitable, Callable, Dict

from .compat import call_async, camel_to_snake
from .constants import ALL_HTTP_METHODS
from .request import Request
from .response import Response

Handler = Callable[[Request, Response, Any], Awaitable[None]]
MethodsParam = Union[List[str], all]


class HandlerDoesNotExist(Exception):
    # Raised to signal that no handler exists for a requested HTTP method.
    pass


class View:
    """This class defines how all HTTP views are represented internally.

    ::: warning
    Views objects should not be created directly. Instead, use
    [from_handler](#from-handler) or [from_obj](#from-obj).
    :::

    HTTP methods are mapped to **handlers**, i.e. methods of a `View` object.

    The following handlers are supported:

    - `.get(req, res, **kwargs)`
    - `.post(req, res, **kwargs)`
    - `.put(req, res, **kwargs)`
    - `.patch(req, res, **kwargs)`
    - `.delete(req, res, **kwargs)`
    - `.head(req, res, **kwargs)`
    - `.options(req, res, **kwargs)`
    - `.handle(req, res, **kwargs)`

    ::: tip
    `.handle()` is special: if defined, it overrides all others.
    :::

    # Attributes

    name (str): the name of the view.
    """

    def __init__(self, name: str):
        self.name = name

    get: Handler
    post: Handler
    put: Handler
    patch: Handler
    delete: Handler
    head: Handler
    options: Handler

    @classmethod
    def create(cls, name: str, docstring: str, handlers: dict) -> "View":
        # Create a view object.
        view = cls(name)
        view.__doc__ = docstring

        # Convert handlers to async if necessary
        for method, handler in handlers.items():
            if not inspect.iscoroutinefunction(handler):
                handler = wraps(handler)(partial(call_async, handler))
                handlers[method] = handler

        # Set head handler if not given but get is given.
        if "get" in handlers and "head" not in handlers:
            handlers["head"] = handlers["get"]

        for method, handler in handlers.items():
            setattr(view, method, handler)

        return view

    def _get_handler(self, req):
        if hasattr(self, "handle"):
            return self.handle
        return getattr(self, req.method.lower())

    async def __call__(self, req, res, **kwargs):
        try:
            handler: Handler = self._get_handler(req)
        except AttributeError as e:
            raise HandlerDoesNotExist from e
        else:
            await handler(req, res, **kwargs)


def from_handler(handler: Handler, methods: MethodsParam = None) -> View:
    """Convert a handler to a `View` instance.

    # Parameters
    handler (function or coroutine function):
        Its name and docstring are copied onto the view.
        It used as a handler for each of the declared `methods`.
        For example, if `methods=["post"]` then the returned `view` object
        will only have a `.post()` handler.
    methods (list of str):
        A list of supported HTTP methods. The `all` built-in can be used
        to support all HTTP methods. Defaults to `["get"]`.

    # Returns
    view (View): a `View` instance.

    # See Also
    - The [constants](./constants.md) module for the list of all HTTP methods.
    """
    if methods is None:
        methods = ["get"]
    if methods is all:
        methods = ["handle"]
    else:
        methods = [m.lower() for m in methods]
    handlers = {method: handler for method in methods}
    return View.create(handler.__name__, handler.__doc__, handlers)


def from_obj(obj: Any) -> View:
    """Convert an object to a `View` instance.

    # Parameters
    obj (any):
        its handlers, snake-cased class name and docstring are copied
        onto the view.

    # Returns
    view (View): a `View` instance.
    """
    handlers = get_handlers(obj)
    name = camel_to_snake(obj.__class__.__name__)
    return View.create(name, obj.__doc__, handlers)


def get_handlers(obj: Any) -> Dict[str, Handler]:
    """Return all `View` handlers declared on an object.

    # Parameters
    obj (any): an object.

    # Returns
    handlers (dict):
        A dict mapping an HTTP method to a handler.
    """
    return {
        method: getattr(obj, method)
        for method in ("handle", *map(str.lower, ALL_HTTP_METHODS))
        if hasattr(obj, method)
    }


def view(methods: MethodsParam = None):
    """Convert the decorated function to a proper `View` object.

    This decorator is a shortcut for [from_handler](#from-handler).
    """
    return partial(from_handler, methods=methods)



from typing import Callable, List, Tuple, Optional

from starlette.middleware.lifespan import LifespanMiddleware

from .app_types import ASGIApp, ASGIAppInstance, Scope

EventHandler = Callable[[], None]


def lifespan_app(scope: Scope) -> ASGIAppInstance:
    # Strict implementation of the ASGI lifespan spec.
    # This is required because the Starlette `LifespanMiddleware`
    # does not send the `complete` responses.

    async def asgi(receive, send):
        message = await receive()
        assert message["type"] == "lifespan.startup"
        await send({"type": "lifespan.startup.complete"})

        message = await receive()
        assert message["type"] == "lifespan.shutdown"
        await send({"type": "lifespan.shutdown.complete"})

    return asgi


class EventsMixin:
    """Allow to register event handlers called during the ASGI lifecycle."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._events: List[Tuple[str, EventHandler]] = []

    def on(self, event: str, handler: Optional[EventHandler] = None):
        """Register an event handler.

        # Parameters
        event (str):
            Either `"startup"` (when the server boots) or `"shutdown"`
            (when the server stops).
        handler (callback, optional):
            The event handler. If not given, this should be used as a
            decorator.

        # Example

        ```python
        @api.on("startup")
        async def init_app():
            pass
        ```
        """

        if handler is not None:
            self._add_event_handler(event, handler)
            return
        else:

            def decorate(f: EventHandler):
                self._add_event_handler(event, f)
                return f

            return decorate

    def _add_event_handler(self, event: str, handler: EventHandler):
        self._events.append((event, handler))

    def _get_lifespan_middleware(self, app: ASGIApp):
        middleware = LifespanMiddleware(app)
        for event, func in self._events:
            middleware.add_event_handler(event, func)
        return middleware

    def handle_lifespan(self, scope: Scope) -> ASGIAppInstance:
        """Create an ASGI application instance to handle `lifespan` messages.

        Registered event handlers will be called as appropriate.

        # Parameters
        scope (dict): an ASGI connection scope.

        # Returns
        app (ASGIAppInstance): an ASGI application instance.

        # See Also
        - [on](#on)
        - [Lifespan protocol](https://asgi.readthedocs.io/en/latest/specs/lifespan.html)
        """
        return self._get_lifespan_middleware(lifespan_app)(scope)
